# -*- coding: utf-8 -*-
"""GPIB Session implementation using linux-gpib or gpib-ctypes.


:copyright: 2015-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import ctypes  # Used for missing bindings not ideal
from bisect import bisect
from typing import Any, Iterator, List, Tuple, Union

from pyvisa import attributes, constants, logger
from pyvisa.constants import ResourceAttribute, StatusCode
from pyvisa.rname import GPIBInstr, GPIBIntfc

from .sessions import Session, UnknownAttribute

try:
    GPIB_CTYPES = True
    from gpib_ctypes import gpib  # typing: ignore
    from gpib_ctypes.Gpib import Gpib  # typing: ignore
    from gpib_ctypes.gpib.gpib import _lib as gpib_lib  # typing: ignore

    try:
        # Add some extra binding not available by default
        extra_funcs = [
            ("ibcac", [ctypes.c_int, ctypes.c_int], ctypes.c_int),
            ("ibgts", [ctypes.c_int, ctypes.c_int], ctypes.c_int),
            ("ibpct", [ctypes.c_int], ctypes.c_int),
        ]
        for name, argtypes, restype in extra_funcs:
            libfunction = gpib_lib[name]
            libfunction.argtypes = argtypes
            libfunction.restype = restype
    except TypeError:
        msg = (
            "gpib_ctypes is installed but could not locate the gpib library.\n"
            "Please manually load it using:\n"
            "  gpib_ctypes.gpib.gpib._load_lib(filename)\n"
            "before importing pyvisa."
        )
        Session.register_unavailable(constants.InterfaceType.gpib, "INSTR", msg)
        Session.register_unavailable(constants.InterfaceType.gpib, "INTFC", msg)
        raise

except ImportError:
    GPIB_CTYPES = False
    try:
        import gpib  # typing: ignore
        from Gpib import Gpib  # typing: ignore
    except ImportError as e:
        msg = (
            "Please install linux-gpib (Linux) or gpib-ctypes (Windows, Linux) "
            "to use this resource type. Note that installing gpib-ctypes will "
            "give you access to a broader range of functionalities.\n%s" % e
        )
        Session.register_unavailable(constants.InterfaceType.gpib, "INSTR", msg)
        Session.register_unavailable(constants.InterfaceType.gpib, "INTFC", msg)
        raise


# patch Gpib to avoid double closing of handles
def _patch_Gpib() -> None:
    if not hasattr(Gpib, "close"):
        _old_del = Gpib.__del__

        def _inner(self):
            _old_del(self)
            self._own = False

        Gpib.__del__ = _inner
        Gpib.close = _inner


_patch_Gpib()


def _find_boards() -> Iterator[Tuple[int, int]]:
    """Find GPIB board addresses."""
    for board in range(16):
        try:
            yield board, gpib.ask(board, 1)
        except gpib.GpibError as e:
            logger.debug("GPIB board %i error in _find_boards(): %s", board, repr(e))


def _find_listeners() -> Iterator[Tuple[int, int, int]]:
    """Find GPIB listeners."""
    for board, boardpad in _find_boards():
        for i in range(31):
            j = 0
            try:
                if boardpad != i and gpib.listener(board, i):
                    yield board, i, j
                elif boardpad != i:
                    for j in range(96, 126):
                        if gpib.listener(board, i, j):
                            yield board, i, j
            except gpib.GpibError as e:
                logger.debug(
                    "GPIB board %i paddr %i saddr %i error in _find_listeners(): %s",
                    board,
                    i,
                    j,
                    repr(e),
                )


def _analyse_lines_value(value: int, line: int):
    """Determine the state of a GPIB line based on the iblines byte.

    Parameters
    ----------
    value : int
        Value returned by iblines.
    line : int
        One of constants.VI_ATTR_GPIB_***_STATE where the *** can be REN, ATN,
        NDAC, or SRQ

    Returns
    -------
    constants.LineState
        State of the line
    StatusCode
        Library StatusCode for the operation.

    """
    if line == constants.VI_ATTR_GPIB_REN_STATE:
        # REN bit valid = 0x10, REN bit value = 0x100
        validity_mask = 0x10
        value_mask = 0x100
    elif line == constants.VI_ATTR_GPIB_ATN_STATE:
        # ATN bit valid = 0x40, ATN bit value = 0x4000
        validity_mask = 0x40
        value_mask = 0x4000
    elif line == constants.VI_ATTR_GPIB_NDAC_STATE:
        # NDAC bit valid = 0x2, NDAC bit value = 0x200
        validity_mask = 0x2
        value_mask = 0x200
    elif line == constants.VI_ATTR_GPIB_SRQ_STATE:
        # SRQ bit valid = 0x20, SRQ bit value = 0x2000
        validity_mask = 0x20
        value_mask = 0x2000

    if not value & validity_mask:
        return constants.LineState.unknown, StatusCode.success
    else:
        if value & value_mask:
            return constants.LineState.asserted, StatusCode.success
        else:
            return constants.LineState.unasserted, StatusCode.success


# linux-gpib timeout constants, in seconds. See GPIBSession._set_timeout.
TIMETABLE = (
    0,
    10e-6,
    30e-6,
    100e-6,
    300e-6,
    1e-3,
    3e-3,
    10e-3,
    30e-3,
    100e-3,
    300e-3,
    1.0,
    3.0,
    10.0,
    30.0,
    100.0,
    300.0,
    1000.0,
)


def convert_gpib_error(
    error: gpib.GpibError, status: int, operation: str
) -> StatusCode:
    """Convert a GPIB error to a VISA StatusCode.

    Parameters
    ----------
    error : gpib.GpibError
        Error to use to determine the proper status code.
    status : int
        Status byte of the GPIB library.
    operation : str
        Name of the operation that caused an exception. Used in logging.

    Returns
    -------
    StatusCode
        Status code matching the GPIB error.

    """
    # First check the imeout condition in the status byte
    if status & 0x4000:
        return StatusCode.error_timeout
    # All other cases are hard errors.
    # In particular linux-gpib simply gives a string we could parse but that
    # feels brittle. As a consequence we only try to be smart when using
    # gpib-ctypes. However in both cases we log the exception at debug level.
    else:
        logger.debug("Failed to %s.", operation, exc_info=error)
        if not GPIB_CTYPES:
            return StatusCode.error_system_error
        if error.code == 1:
            return StatusCode.error_not_cic
        elif error.code == 2:
            return StatusCode.error_no_listeners
        elif error.code == 4:
            return StatusCode.error_invalid_mode
        elif error.code == 11:
            return StatusCode.error_nonsupported_operation
        elif error.code == 1:
            return StatusCode.error_not_cic
        elif error.code == 21:
            return StatusCode.error_resource_locked
        else:
            return StatusCode.error_system_error


def convert_gpib_status(status: int) -> StatusCode:
    if status & 0x4000:
        return StatusCode.error_timeout
    elif status & 0x8000:
        return StatusCode.error_system_error
    else:
        return StatusCode.success


class _GPIBCommon(Session):
    """Common base class for GPIB sessions.

    Both INSTR and INTFC resources share the following attributes:
    - VI_ATTR_INTF_TYPE
    - VI_ATTR_TMO_VALUE
    - VI_ATTR_INTF_INST_NAME
    - VI_ATTR_INTF_NUM
    - VI_ATTR_DMA_ALLOW_EN
    - VI_ATTR_SEND_END_EN
    - VI_ATTR_TERMCHAR
    - VI_ATTR_TERM_CHAR_EN
    - VI_ATTR_RD_BUF_OPER_MODE
    - VI_ATTR_WR_BUF_OPER_MODE
    - VI_ATTR_FILE_APPEND_EN
    - VI_ATTR_GPIB_PRIMARY_ADDR
    - VI_ATTR_GPIB_SECONDARY_ADDR
    - VI_ATTR_GPIB_REN_STATE

    """

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: Union[GPIBIntfc, GPIBInstr]

    #: Bus wide controller.
    controller: Gpib

    @classmethod
    def get_low_level_info(cls) -> str:
        try:
            ver = gpib.version()
        except AttributeError:
            ver = "< 4.0"

        return "via Linux GPIB (%s)" % ver

    def after_parsing(self) -> None:
        minor = int(self.parsed.board)
        # Secondary address (SAD) values should be in the range 96 to 126,
        # 0 means the SAD is disabled.
        sad = 0
        timeout = 13
        send_eoi = 1
        eos_mode = 0
        self.interface = None
        if isinstance(self.parsed, GPIBInstr):
            pad = int(self.parsed.primary_address)
            if self.parsed.secondary_address is not None:
                sad = int(self.parsed.secondary_address) + 0x60
            # Used to talk to a specific resource
            self.interface = Gpib(
                name=minor,
                pad=pad,
                sad=sad,
                timeout=timeout,
                send_eoi=send_eoi,
                eos_mode=eos_mode,
            )
        # Bus wide operation
        self.controller = Gpib(name=minor)

        # Force timeout setting to interface
        self.set_attribute(
            constants.ResourceAttribute.timeout_value,
            attributes.AttributesByID[constants.VI_ATTR_TMO_VALUE].default,
        )

        for name in ("TERMCHAR", "TERMCHAR_EN"):
            attribute = getattr(constants, "VI_ATTR_" + name)
            self.attrs[attribute] = attributes.AttributesByID[attribute].default

    def _get_timeout(
        self, attribute: constants.ResourceAttribute
    ) -> Tuple[int, StatusCode]:
        if self.interface:
            # 0x3 is the hexadecimal reference to the IbaTMO (timeout) configuration
            # option in linux-gpib.
            gpib_timeout = self.interface.ask(3)
            if gpib_timeout and gpib_timeout < len(TIMETABLE):
                self.timeout = TIMETABLE[gpib_timeout]
            else:
                # value is 0 or out of range -> infinite
                self.timeout = None
        return super(_GPIBCommon, self)._get_timeout(attribute)

    def _set_timeout(self, attribute: constants.ResourceAttribute, value: int):
        """Set the timeout value.

        linux-gpib only supports 18 discrete timeout values. If a timeout
        value other than these is requested, it will be rounded up to the closest
        available value. Values greater than the largest available timout value
        will instead be rounded down. The available timeout values are:
        0   Never timeout.
        1   10 microseconds
        2   30 microseconds
        3   100 microseconds
        4   300 microseconds
        5   1 millisecond
        6   3 milliseconds
        7   10 milliseconds
        8   30 milliseconds
        9   100 milliseconds
        10  300 milliseconds
        11  1 second
        12  3 seconds
        13  10 seconds
        14  30 seconds
        15  100 seconds
        16  300 seconds
        17  1000 seconds

        """
        status = super(_GPIBCommon, self)._set_timeout(attribute, value)
        # Inspect the result of setting the value to decide how to translate the result
        # on the interface.
        if self.interface:
            if self.timeout is None:
                gpib_timeout = 0
            else:
                # round up only values that are higher by 0.1% than discrete values
                gpib_timeout = min(bisect(TIMETABLE, 0.999 * self.timeout), 17)
                self.timeout = TIMETABLE[gpib_timeout]
            self.interface.timeout(gpib_timeout)
        return status

    def close(self) -> StatusCode:
        if self.interface:
            self.interface.close()
        self.controller.close()
        return StatusCode.success

    def read(self, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        ----------
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read from the interface.
        StatusCode
            Return value of the library call.

        """
        # INTFC don't have an interface so use the controller
        ifc = self.interface or self.controller

        # END 0x2000
        checker = lambda current: ifc.ibsta() & 0x2000  # noqa: E731

        reader = lambda: ifc.read(count)  # noqa: E731

        return self._read(reader, count, checker, False, None, False, gpib.GpibError)

    def write(self, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        logger.debug("GPIB.write %r" % data)

        # INTFC don't have an interface so use the controller
        ifc = self.interface or self.controller

        try:
            ifc.write(data)
            count = ifc.ibcnt()  # number of bytes transmitted

            return count, StatusCode.success
        except gpib.GpibError as e:
            return 0, convert_gpib_error(e, ifc.ibsta(), "write")

    def gpib_control_ren(self, mode: constants.RENLineOperation) -> StatusCode:
        """Controls the state of the GPIB Remote Enable (REN) interface line.

        Optionally the remote/local state of the device is also controlled.

        Corresponds to viGpibControlREN function of the VISA library.

        Parameters
        ----------
        mode : constants.RENLineOperation
            Specifies the state of the REN line and optionally the device
            remote/local state.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        if isinstance(self.parsed, GPIBIntfc):
            if mode not in (
                constants.VI_GPIB_REN_ASSERT,
                constants.VI_GPIB_REN_DEASSERT,
                constants.VI_GPIB_REN_ASSERT_LLO,
            ):
                return constants.StatusCode.error_nonsupported_operation

        # INTFC don't have an interface so use the controller
        ifc = self.interface or self.controller
        try:
            if mode == constants.VI_GPIB_REN_DEASSERT_GTL:
                # Send GTL command byte (cf linux-gpib documentation)
                ifc.command(chr(1))
            if mode in (
                constants.VI_GPIB_REN_DEASSERT,
                constants.VI_GPIB_REN_DEASSERT_GTL,
            ):
                self.controller.remote_enable(0)

            if mode == constants.VI_GPIB_REN_ASSERT_LLO:
                # LLO
                ifc.command(b"0x11")
            elif mode == constants.VI_GPIB_REN_ADDRESS_GTL:
                # GTL
                ifc.command(b"0x1")
            elif mode == constants.VI_GPIB_REN_ASSERT_ADDRESS_LLO:
                pass
            elif mode in (
                constants.VI_GPIB_REN_ASSERT,
                constants.VI_GPIB_REN_ASSERT_ADDRESS,
            ):
                ifc.remote_enable(1)
                if (
                    isinstance(self.parsed, GPIBInstr)
                    and mode == constants.VI_GPIB_REN_ASSERT_ADDRESS
                ):
                    # 0 for the secondary address means don't use it
                    ifc.listener(
                        self.parsed.primary_address, self.parsed.secondary_address
                    )
        except gpib.GpibError as e:
            return convert_gpib_error(e, self.interface.ibsta(), "perform control REN")

        return constants.StatusCode.success

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state query is made

        Returns
        -------
        Any
            The state of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_INTF_INST_NAME RO
        # - VI_ATTR_DMA_ALLOW_EN RW
        # - VI_ATTR_RD_BUF_OPER_MODE RW
        # - VI_ATTR_WR_BUF_OPER_MODE RW
        # - VI_ATTR_FILE_APPEND_EN RW

        # INTFC don't have an interface so use the controller
        ifc = self.interface or self.controller

        if attribute == ResourceAttribute.gpib_primary_address:
            # IbaPAD 0x1
            return ifc.ask(1), StatusCode.success

        elif attribute == ResourceAttribute.gpib_secondary_address:
            # IbaSAD 0x2
            # Remove 0x60 because National Instruments.
            _ = ifc.ask(2)
            if ifc.ask(2):
                return ifc.ask(2) - 96, StatusCode.success
            else:
                return constants.VI_NO_SEC_ADDR, StatusCode.success

        elif attribute == ResourceAttribute.gpib_ren_state:
            try:
                lines = self.controller.lines()
                return _analyse_lines_value(lines, attribute)
            except AttributeError:
                # some versions of linux-gpib do not expose Gpib.lines()
                return constants.VI_STATE_UNKNOWN, StatusCode.success

        elif attribute == ResourceAttribute.send_end_enabled:
            # Do not use IbaEndBitIsNormal 0x1a which relates to EOI on read()
            # not write(). see issue #196
            # IbcEOT 0x4
            if ifc.ask(4):
                return constants.VI_TRUE, StatusCode.success
            else:
                return constants.VI_FALSE, StatusCode.success

        elif attribute == ResourceAttribute.interface_number:
            # IbaBNA 0x200
            return ifc.ask(512), StatusCode.success

        elif attribute == ResourceAttribute.interface_type:
            return constants.InterfaceType.gpib, StatusCode.success

        raise UnknownAttribute(attribute)

    def _set_attribute(
        self, attribute: constants.ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_DMA_ALLOW_EN RW
        # - VI_ATTR_RD_BUF_OPER_MODE RW
        # - VI_ATTR_WR_BUF_OPER_MODE RW
        # - VI_ATTR_FILE_APPEND_EN RW

        # INTFC don't have an interface so use the controller
        ifc = self.interface or self.controller

        if attribute == ResourceAttribute.gpib_readdress_enabled:
            # IbcREADDR 0x6
            # Setting has no effect in linux-gpib.
            if isinstance(attribute_state, int):
                ifc.config(6, attribute_state)
                return StatusCode.success
            else:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == ResourceAttribute.gpib_primary_address:
            # IbcPAD 0x1
            if isinstance(attribute_state, int) and 0 <= attribute_state <= 30:
                ifc.config(1, attribute_state)
                return StatusCode.success
            else:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == ResourceAttribute.gpib_secondary_address:
            # IbcSAD 0x2
            # Add 0x60 because National Instruments.
            if isinstance(attribute_state, int) and 0 <= attribute_state <= 30:
                if ifc.ask(2):
                    ifc.config(2, attribute_state + 96)
                    return StatusCode.success
                else:
                    return StatusCode.error_nonsupported_attribute
            else:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == ResourceAttribute.gpib_unadress_enable:
            # IbcUnAddr 0x1b
            try:
                ifc.config(27, attribute_state)
                return StatusCode.success
            except gpib.GpibError:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == ResourceAttribute.send_end_enabled:
            # Do not use IbaEndBitIsNormal 0x1a which relates to EOI on read()
            # not write(). see issue #196
            # IbcEOT 0x4
            if isinstance(attribute_state, int):
                ifc.config(4, attribute_state)
                return StatusCode.success
            else:
                return StatusCode.error_nonsupported_attribute_state

        raise UnknownAttribute(attribute)


# TODO: Check secondary addresses.
@Session.register(constants.InterfaceType.gpib, "INSTR")
class GPIBSession(_GPIBCommon):
    """A GPIB Session that uses linux-gpib to do the low level communication."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: GPIBInstr

    @staticmethod
    def list_resources() -> List[str]:
        return [
            "GPIB%d::%d::INSTR" % (board, pad)
            if sad == 0
            else "GPIB%d::%d::%d::INSTR" % (board, pad, sad - 0x60)
            for board, pad, sad in _find_listeners()
        ]

    def clear(self) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        logger.debug("GPIB.device clear")
        try:
            self.interface.clear()
            return StatusCode.success
        except gpib.GpibError as e:
            return convert_gpib_error(e, self.interface.ibsta(), "clear")

    def assert_trigger(self, protocol: constants.TriggerProtocol) -> StatusCode:
        """Asserts hardware trigger.

        Parameters
        ----------
        protocol : constants.TriggerProtocol
            Triggering protocol to use.
            Only supports constants.TriggerProtocol.default

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        logger.debug("GPIB.device assert hardware trigger")

        try:
            if protocol == constants.VI_TRIG_PROT_DEFAULT:
                self.interface.trigger()
                return StatusCode.success
            else:
                return StatusCode.error_nonsupported_operation
        except gpib.GpibError as e:
            return convert_gpib_error(e, self.interface.ibsta(), "assert trigger")

    def read_stb(self) -> Tuple[int, StatusCode]:
        """Read the device status byte."""
        try:
            return self.interface.serial_poll(), StatusCode.success
        except gpib.GpibError as e:
            return 0, convert_gpib_error(e, self.interface.ibsta(), "read STB")

    def _get_attribute(
        self, attribute: constants.ResourceAttribute
    ) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes. GPIB::INSTR have the
        following specific attributes:

        - VI_ATTR_TRIG_ID
        - VI_ATTR_IO_PROT
        - VI_ATTR_SUPPRESS_END_EN
        - VI_ATTR_GPIB_READDR_EN
        - VI_ATTR_GPIB_UNADDR_EN

        Parameters
        ----------
        attribute : constants.ResourceAttribute
            Resource attribute for which the state query is made

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_TRIG_ID RW or RO see specs
        # - VI_ATTR_IO_PROT RW
        # - VI_ATTR_SUPPRESS_END_EN RW
        ifc = self.interface

        if attribute == constants.VI_ATTR_GPIB_READDR_EN:
            # IbaREADDR 0x6
            # Setting has no effect in linux-gpib.
            return ifc.ask(6), StatusCode.success

        elif attribute == constants.VI_ATTR_GPIB_UNADDR_EN:
            # IbaUnAddr 0x1b
            if ifc.ask(27):
                return constants.VI_TRUE, StatusCode.success
            else:
                return constants.VI_FALSE, StatusCode.success

        return super(GPIBSession, self)._get_attribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state is to be modified. (Attributes.*)
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_TRIG_ID RW or RO see specs
        # - VI_ATTR_IO_PROT RW
        # - VI_ATTR_SUPPRESS_END_EN RW
        ifc = self.interface

        if attribute == constants.VI_ATTR_GPIB_READDR_EN:
            # IbcREADDR 0x6
            # Setting has no effect in linux-gpib.
            if isinstance(attribute_state, int):
                ifc.config(6, attribute_state)
                return StatusCode.success
            else:
                return StatusCode.error_nonsupported_attribute_state

        elif attribute == constants.VI_ATTR_GPIB_UNADDR_EN:
            # IbcUnAddr 0x1b
            try:
                ifc.config(27, attribute_state)
                return StatusCode.success
            except gpib.GpibError:
                return StatusCode.error_nonsupported_attribute_state

        return super(GPIBSession, self)._set_attribute(attribute, attribute_state)


@Session.register(constants.InterfaceType.gpib, "INTFC")
class GPIBInterface(_GPIBCommon):
    """A GPIB Interface that uses linux-gpib to do the low level communication."""

    # Override parsed to take into account the fact that this class is only used
    # for a specific kind of resource
    parsed: GPIBIntfc

    @staticmethod
    def list_resources() -> List[str]:
        return ["GPIB%d::INTFC" % board for board, pad in _find_boards()]

    def gpib_command(self, command_bytes: bytes) -> Tuple[int, StatusCode]:
        """Write GPIB command byte on the bus.

        Corresponds to viGpibCommand function of the VISA library.
        See: https://linux-gpib.sourceforge.io/doc_html/gpib-protocol.html#REFERENCE-COMMAND-BYTES

        Parameters
        ----------
        command_bytes : bytes
            Command bytes to send

        Returns
        -------
        int
            Number of written bytes,
        StatusCode
            Return value of the library call.

        """
        try:
            return self.controller.command(command_bytes), StatusCode.success
        except gpib.GpibError as e:
            return 0, convert_gpib_error(e, self.controller.ibsta(), "gpib command")

    def gpib_send_ifc(self) -> StatusCode:
        """Pulse the interface clear line (IFC) for at least 100 microseconds.

        Corresponds to viGpibSendIFC function of the VISA library.

        """
        logger.debug("GPIB.interface clear")
        try:
            self.controller.interface_clear()
            return StatusCode.success
        except gpib.GpibError as e:
            return convert_gpib_error(e, self.controller.ibsta(), "send IFC")

    def gpib_control_atn(self, mode: constants.ATNLineOperation) -> StatusCode:
        """Specifies the state of the ATN line and the local active controller state.

        Corresponds to viGpibControlATN function of the VISA library.

        Parameters
        ----------
        mode : constants.ATNLineOperation
            Specifies the state of the ATN line and optionally the local active
             controller state.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        logger.debug("GPIB.control atn")
        if mode == constants.VI_GPIB_ATN_ASSERT:
            status = gpib_lib.ibcac(self.controller.id, 0)
        elif mode == constants.VI_GPIB_ATN_DEASSERT:
            status = gpib_lib.ibgts(self.controller.id, 0)
        elif mode == constants.VI_GPIB_ATN_ASSERT_IMMEDIATE:
            # Asynchronous assertion (the name is counter intuitive)
            status = gpib_lib.ibcac(self.controller.id, 1)
        elif mode == constants.VI_GPIB_ATN_DEASSERT_HANDSHAKE:
            status = gpib_lib.ibgts(self.controller.id, 1)
        else:
            return constants.StatusCode.error_invalid_mode
        return convert_gpib_status(status)

    def gpib_pass_control(
        self, primary_address: int, secondary_address: int
    ) -> StatusCode:
        """Tell a GPIB device to become controller in charge (CIC).

        Corresponds to viGpibPassControl function of the VISA library.

        Parameters
        ----------
        primary_address : int
            Primary address of the GPIB device to which you want to pass control.
        secondary_address : int
            Secondary address of the targeted GPIB device.
            If the targeted device does not have a secondary address,
            this parameter should contain the value Constants.VI_NO_SEC_ADDR.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        # ibpct need to get the device id matching the primary and secondary address
        logger.debug("GPIB.pass control")
        try:
            did = gpib.dev(self.parsed.board, primary_address, secondary_address)
        except gpib.GpibError:
            logger.exception(
                "Failed to get id for %s, %d", primary_address, secondary_address
            )
            return StatusCode.error_resource_not_found

        status = gpib_lib.ibpct(did)
        return convert_gpib_status(status)

    def _get_attribute(self, attribute: ResourceAttribute) -> Tuple[Any, StatusCode]:
        """Get the value for a given VISA attribute for this session.

        Use to implement custom logic for attributes. GPIB::INTFC have the
        following specific attributes:

        - VI_ATTR_DEV_STATUS_BYTE
        - VI_ATTR_GPIB_ATN_STATE
        - VI_ATTR_GPIB_NDAC_STATE
        - VI_ATTR_GPIB_SRQ_STATE
        - VI_ATTR_GPIB_CIC_STATE
        - VI_ATTR_GPIB_SYS_CNTRL_STATE
        - VI_ATTR_GPIB_HS488_CBL_LEN
        - VI_ATTR_GPIB_ADDR_STATE

        Parameters
        ----------
        attribute: ResourceAttribute
            Resource attribute for which the state query is made.

        Returns
        -------
        Any
            The state of the queried attribute for a specified resource,
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_DEV_STATUS_BYTE RW
        # - VI_ATTR_GPIB_SYS_CNTRL_STATE RW
        # - VI_ATTR_GPIB_HS488_CBL_LEN RO
        # - VI_ATTR_GPIB_ADDR_STATE RO
        ifc = self.controller

        if attribute == constants.VI_ATTR_GPIB_CIC_STATE:
            # ibsta CIC = 0x0020
            if ifc.ibsta() & 0x0020:
                return constants.VI_TRUE, StatusCode.success
            else:
                return constants.VI_FALSE, StatusCode.success

        elif attribute in (
            constants.VI_ATTR_GPIB_ATN_STATE,
            constants.VI_ATTR_GPIB_NDAC_STATE,
            constants.VI_ATTR_GPIB_SRQ_STATE,
        ):
            try:
                lines = ifc.lines()
                return _analyse_lines_value(lines, attribute)
            except AttributeError:
                # some versions of linux-gpib do not expose Gpib.lines()
                return constants.VI_STATE_UNKNOWN, StatusCode.success

        return super()._get_attribute(attribute)

    def _set_attribute(
        self, attribute: ResourceAttribute, attribute_state: Any
    ) -> StatusCode:
        """Sets the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        attribute : ResourceAttribute
            Attribute for which the state is to be modified.
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        # TODO implement the following attributes
        # - VI_ATTR_GPIB_SYS_CNTRL_STATE
        # - VI_ATTR_DEV_STATUS_BYTE

        # INTFC don't have an interface so use the controller
        _ = self.controller

        return super()._set_attribute(attribute, attribute_state)



# -*- coding: utf-8 -*-
"""Highlevel wrapper of the VISA Library.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import random
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, cast

from pyvisa import constants, highlevel, rname
from pyvisa.constants import StatusCode
from pyvisa.typing import VISAEventContext, VISARMSession, VISASession
from pyvisa.util import DebugInfo, LibraryPath

from . import sessions
from .common import logger


class PyVisaLibrary(highlevel.VisaLibraryBase):
    """A pure Python backend for PyVISA.

    The object is basically a dispatcher with some common functions implemented.

    When a new resource object is requested to pyvisa, the library creates a
    Session object (that knows how to perform low-level communication operations)
    associated with a session handle (a number, usually refered just as session).

    A call to a library function is handled by PyVisaLibrary if it involves a
    resource agnostic function or dispatched to the correct session object
    (obtained from the session id).

    Importantly, the user is unaware of this. PyVisaLibrary behaves for
    the user just as NIVisaLibrary.

    """

    #: Live session object identified by a randon session ID
    sessions: Dict[int, sessions.Session]

    # Try to import packages implementing lower level functionality.
    try:
        from .serial import SerialSession

        logger.debug("SerialSession was correctly imported.")
    except Exception as e:
        logger.debug("SerialSession was not imported %s." % e)

    try:
        from .usb import USBRawSession, USBSession

        logger.debug("USBSession and USBRawSession were correctly imported.")
    except Exception as e:
        logger.debug("USBSession and USBRawSession were not imported %s." % e)

    try:
        from .tcpip import TCPIPInstrSession, TCPIPSocketSession

        logger.debug("TCPIPSession was correctly imported.")
    except Exception as e:
        logger.debug("TCPIPSession was not imported %s." % e)

    try:
        from .gpib import GPIBSession

        logger.debug("GPIBSession was correctly imported.")
    except Exception as e:
        logger.debug("GPIBSession was not imported %s." % e)

    @staticmethod
    def get_library_paths() -> Iterable[LibraryPath]:
        """List a dummy library path to allow to create the library."""
        return (LibraryPath("py"),)

    @staticmethod
    def get_debug_info() -> DebugInfo:
        """Return a list of lines with backend info."""
        from . import __version__

        d: OrderedDict[str, Union[str, List[str], Dict[str, str]]] = OrderedDict()
        d["Version"] = "%s" % __version__

        for key, val in sessions.Session.iter_valid_session_classes():
            key_name = "%s %s" % (key[0].name.upper(), key[1])
            d[key_name] = "Available " + val.get_low_level_info()

        for key, issue in sessions.Session.iter_session_classes_issues():
            key_name = "%s %s" % (key[0].name.upper(), key[1])
            d[key_name] = issue.split("\n")

        return d

    def _init(self) -> None:
        """Custom initialization code."""
        # Map session handle to session object.
        self.sessions = {}

    def _register(self, obj: object) -> VISASession:
        """Creates a random but unique session handle for a session object.

        Register it in the sessions dictionary and return the value.

        """
        session = None

        while session is None or session in self.sessions:
            session = random.randint(1000000, 9999999)

        self.sessions[session] = obj
        return session

    def open(
        self,
        session: VISARMSession,
        resource_name: str,
        access_mode: constants.AccessModes = constants.AccessModes.no_lock,
        open_timeout: Optional[int] = constants.VI_TMO_IMMEDIATE,
    ) -> Tuple[VISASession, StatusCode]:
        """Opens a session to the specified resource.

        Corresponds to viOpen function of the VISA library.

        Parameters
        ----------
        session : VISARMSession
            Resource Manager session (should always be a session returned from
            open_default_resource_manager()).
        resource_name : str
            Unique symbolic name of a resource.
        access_mode : constants.AccessModes, optional
            Specifies the mode by which the resource is to be accessed.
        open_timeout : int
            Specifies the maximum time period (in milliseconds) that this
            operation waits before returning an error. constants.VI_TMO_IMMEDIATE
            and constants.VI_TMO_INFINITE are used as min and max.

        Returns
        -------
        VISASession
            Unique logical identifier reference to a session
        StatusCode
            Return value of the library call.

        """
        try:
            open_timeout = None if open_timeout is None else int(open_timeout)
        except ValueError:
            raise ValueError(
                "open_timeout (%r) must be an integer (or compatible type)"
                % open_timeout
            )

        try:
            parsed = rname.parse_resource_name(resource_name)
        except rname.InvalidResourceName:
            return (
                VISASession(0),
                self.handle_return_value(None, StatusCode.error_invalid_resource_name),
            )

        cls = sessions.Session.get_session_class(
            parsed.interface_type_const, parsed.resource_class
        )

        sess = cls(session, resource_name, parsed, open_timeout)

        return self._register(sess), StatusCode.success

    def clear(self, session: VISASession) -> StatusCode:
        """Clears a device.

        Corresponds to viClear function of the VISA library.

        Parameters
        ----------
        session : typing.VISASession
            Unique logical identifier to a session.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)
        return self.handle_return_value(session, sess.clear())

    def flush(
        self, session: VISASession, mask: constants.BufferOperation
    ) -> StatusCode:
        """Flush the specified buffers.

        The buffers can be associated with formatted I/O operations and/or
        serial communication.

        Corresponds to viFlush function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        mask : constants.BufferOperation
            Specifies the action to be taken with flushing the buffer.
            The values can be combined using the | operator. However multiple
            operations on a single buffer cannot be combined.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)
        return self.handle_return_value(session, sess.flush(mask))

    def gpib_command(
        self, session: VISASession, command_byte: bytes
    ) -> Tuple[int, StatusCode]:
        """Write GPIB command bytes on the bus.

        Corresponds to viGpibCommand function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        command_byte : bytes
            Data to write.

        Returns
        -------
        int
            Number of written bytes
        StatusCode
            Return value of the library call.

        """
        try:
            written, st = self.sessions[session].gpib_command(command_byte)
            return written, self.handle_return_value(session, st)
        except KeyError:
            return 0, self.handle_return_value(session, StatusCode.error_invalid_object)

    def assert_trigger(
        self, session: VISASession, protocol: constants.TriggerProtocol
    ) -> StatusCode:
        """Assert software or hardware trigger.

        Corresponds to viAssertTrigger function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        protocol : constants.TriggerProtocol
            Trigger protocol to use during assertion.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session, self.sessions[session].assert_trigger(protocol)
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def gpib_send_ifc(self, session: VISASession) -> StatusCode:
        """Pulse the interface clear line (IFC) for at least 100 microseconds.

        Corresponds to viGpibSendIFC function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session, self.sessions[session].gpib_send_ifc()
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def gpib_control_ren(
        self, session: VISASession, mode: constants.RENLineOperation
    ) -> StatusCode:
        """Controls the state of the GPIB Remote Enable (REN) interface line.

        Optionally the remote/local state of the device can also be set.

        Corresponds to viGpibControlREN function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        mode : constants.RENLineOperation
            State of the REN line and optionally the device remote/local state.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session, self.sessions[session].gpib_control_ren(mode)
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def gpib_control_atn(
        self, session: VISASession, mode: constants.ATNLineOperation
    ) -> StatusCode:
        """Specifies the state of the ATN line and the local active controller state.

        Corresponds to viGpibControlATN function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        mode : constants.ATNLineOperation
            State of the ATN line and optionally the local active controller state.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session, self.sessions[session].gpib_control_atn(mode)
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def gpib_pass_control(
        self, session: VISASession, primary_address: int, secondary_address: int
    ) -> StatusCode:
        """Tell a GPIB device to become controller in charge (CIC).

        Corresponds to viGpibPassControl function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        primary_address : int
            Primary address of the GPIB device to which you want to pass control.
        secondary_address : int
            Secondary address of the targeted GPIB device.
            If the targeted device does not have a secondary address, this parameter
            should contain the value Constants.VI_NO_SEC_ADDR.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session,
                self.sessions[session].gpib_pass_control(
                    primary_address, secondary_address
                ),
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def read_stb(self, session: VISASession) -> Tuple[int, StatusCode]:
        """Reads a status byte of the service request.

        Corresponds to viReadSTB function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.

        Returns
        -------
        int
            Service request status byte
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return 0, self.handle_return_value(session, StatusCode.error_invalid_object)
        stb, status_code = sess.read_stb()
        return stb, self.handle_return_value(session, status_code)

    def close(
        self, session: Union[VISASession, VISAEventContext, VISARMSession]
    ) -> StatusCode:
        """Closes the specified session, event, or find list.

        Corresponds to viClose function of the VISA library.

        Parameters
        ---------
        session : Union[VISASession, VISAEventContext, VISARMSession]
            Unique logical identifier to a session, event, resource manager.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
            # The RM session directly references the library.
            if sess is not self:
                return self.handle_return_value(session, sess.close())
            else:
                return self.handle_return_value(session, StatusCode.success)
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def open_default_resource_manager(self) -> Tuple[VISARMSession, StatusCode]:
        """This function returns a session to the Default Resource Manager resource.

        Corresponds to viOpenDefaultRM function of the VISA library.

        Returns
        -------
        VISARMSession
            Unique logical identifier to a Default Resource Manager session
        StatusCode
            Return value of the library call.

        """
        return (
            cast(VISARMSession, self._register(self)),
            self.handle_return_value(None, StatusCode.success),
        )

    def list_resources(
        self, session: VISARMSession, query: str = "?*::INSTR"
    ) -> Tuple[str, ...]:
        """Return a tuple of all connected devices matching query.

        Parameters
        ----------
        session : VISARMSession
            Unique logical identifier to the resource manager session.
        query : str
            Regular expression used to match devices.

        Returns
        -------
        Tuple[str, ...]
            Resource names of all the connected devices matching the query.

        """
        # For each session type, ask for the list of connected resources and
        # merge them into a single list.
        # HINT: the cast should not be necessary here
        resources: List[str] = []
        for key, st in sessions.Session.iter_valid_session_classes():
            resources += st.list_resources()

        return rname.filter(resources, query)

    def read(self, session: VISASession, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data from device or interface synchronously.

        Corresponds to viRead function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Date read
        StatusCode
            Return value of the library call.

        """
        # from the session handle, dispatch to the read method of the session object.
        try:
            data, status_code = self.sessions[session].read(count)
        except KeyError:
            return (
                b"",
                self.handle_return_value(session, StatusCode.error_invalid_object),
            )

        return data, self.handle_return_value(session, status_code)

    def write(self, session: VISASession, data: bytes) -> Tuple[int, StatusCode]:
        """Write data to device or interface synchronously.

        Corresponds to viWrite function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        data : bytes
            Data to be written.

        Returns
        -------
        int
            Number of bytes actually transferred
        StatusCode
            Return value of the library call.

        """
        # from the session handle, dispatch to the write method of the session object.
        try:
            written, status_code = self.sessions[session].write(data)
        except KeyError:
            return 0, self.handle_return_value(session, StatusCode.error_invalid_object)

        return written, self.handle_return_value(session, status_code)

    def buffer_read(self, session: VISASession, count: int) -> Tuple[bytes, StatusCode]:
        """Reads data through the use of a formatted I/O read buffer.

        The data can be read from a device or an interface.

        Corresponds to viBufRead function of the VISA library.

        Parameters
        ----------
        session : VISASession\
            Unique logical identifier to a session.
        count : int
            Number of bytes to be read.

        Returns
        -------
        bytes
            Data read
        StatusCode
            Return value of the library call.

        """
        return self.read(session, count)

    def buffer_write(self, session: VISASession, data: bytes) -> Tuple[int, StatusCode]:
        """Writes data to a formatted I/O write buffer synchronously.

        Corresponds to viBufWrite function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        data : bytes
            Data to be written.

        Returns
        -------
        int
            number of written bytes
        StatusCode
            return value of the library call.

        """
        return self.write(session, data)

    def get_attribute(
        self,
        session: Union[VISASession, VISAEventContext, VISARMSession],
        attribute: Union[constants.ResourceAttribute, constants.EventAttribute],
    ) -> Tuple[Any, StatusCode]:
        """Retrieves the state of an attribute.

        Corresponds to viGetAttribute function of the VISA library.

        Parameters
        ----------
        session : Union[VISASession, VISAEventContext]
            Unique logical identifier to a session, event, or find list.
        attribute : Union[constants.ResourceAttribute, constants.EventAttribute]
            Resource or event attribute for which the state query is made.

        Returns
        -------
        Any
            State of the queried attribute for a specified resource
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return (
                None,
                self.handle_return_value(session, StatusCode.error_invalid_object),
            )

        state, status_code = sess.get_attribute(
            cast(constants.ResourceAttribute, attribute)
        )
        return state, self.handle_return_value(session, status_code)

    def set_attribute(
        self,
        session: VISASession,
        attribute: constants.ResourceAttribute,
        attribute_state: Any,
    ) -> StatusCode:
        """Set the state of an attribute.

        Corresponds to viSetAttribute function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        attribute : constants.ResourceAttribute
            Attribute for which the state is to be modified.
        attribute_state : Any
            The state of the attribute to be set for the specified object.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            return self.handle_return_value(
                session,
                self.sessions[session].set_attribute(attribute, attribute_state),
            )
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)

    def lock(
        self,
        session: VISASession,
        lock_type: constants.Lock,
        timeout: int,
        requested_key: Optional[str] = None,
    ) -> Tuple[str, StatusCode]:
        """Establishes an access mode to the specified resources.

        Corresponds to viLock function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        lock_type : constants.Lock
            Specifies the type of lock requested.
        timeout : int
            Absolute time period (in milliseconds) that a resource waits to get
            unlocked by the locking session before returning an error.
        requested_key : Optional[str], optional
            Requested locking key in the case of a shared lock. For an exclusive
            lock it should be None.

        Returns
        -------
        str
            Key that can then be passed to other sessions to share the lock, or
            None for an exclusive lock.
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return (
                "",
                self.handle_return_value(session, StatusCode.error_invalid_object),
            )
        key, status_code = sess.lock(lock_type, timeout, requested_key)
        return key, self.handle_return_value(session, status_code)

    def unlock(self, session: VISASession) -> StatusCode:
        """Relinquish a lock for the specified resource.

        Corresponds to viUnlock function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        try:
            sess = self.sessions[session]
        except KeyError:
            return self.handle_return_value(session, StatusCode.error_invalid_object)
        return self.handle_return_value(session, sess.unlock())

    def disable_event(
        self,
        session: VISASession,
        event_type: constants.EventType,
        mechanism: constants.EventMechanism,
    ) -> StatusCode:
        """Disable notification for an event type(s) via the specified mechanism(s).

        Corresponds to viDisableEvent function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        event_type : constants.EventType
            Event type.
        mechanism : constants.EventMechanism
            Event handling mechanisms to be disabled.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonimplemented_operation

    def discard_events(
        self,
        session: VISASession,
        event_type: constants.EventType,
        mechanism: constants.EventMechanism,
    ) -> StatusCode:
        """Discard event occurrences for a given type and mechanisms in a session.

        Corresponds to viDiscardEvents function of the VISA library.

        Parameters
        ----------
        session : VISASession
            Unique logical identifier to a session.
        event_type : constans.EventType
            Logical event identifier.
        mechanism : constants.EventMechanism
            Specifies event handling mechanisms to be discarded.

        Returns
        -------
        StatusCode
            Return value of the library call.

        """
        return StatusCode.error_nonimplemented_operation



# -*- coding: utf-8 -*-
"""Implements protocols on top of lower level libraries to talk to instruments.


:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""



# -*- coding: utf-8 -*-
"""Implements Session to control USBTMC instruments

Loosely based on PyUSBTMC:python module to handle USB-TMC(Test and
Measurement class)　devices. by Noboru Yamamot, Accl. Lab, KEK, JAPAN

This file is an offspring of the Lantz Project.

:copyright: 2014-2020 by PyVISA-py Authors, see AUTHORS for more details.
:license: MIT, see LICENSE for more details.

"""

import enum
import struct
import time
import warnings
from collections import namedtuple

import usb

from .usbutil import find_devices, find_endpoint, find_interfaces, usb_find_desc


class MsgID(enum.IntEnum):
    """From USB-TMC table2"""

    dev_dep_msg_out = 1
    request_dev_dep_msg_in = 2
    dev_dep_msg_in = 2
    vendor_specific_out = 126
    request_vendor_specific_in = 127
    vendor_specific_in = 127

    # USB488
    trigger = 128


class Request(enum.IntEnum):
    initiate_abort_bulk_out = 1
    check_abort_bulk_out_status = 2
    initiate_abort_bulk_in = 3
    check_abort_bulk_in_status = 4
    initiate_clear = 5
    check_clear_status = 6
    get_capabilities = 7
    indicator_pulse = 64

    # USB488
    read_status_byte = 128
    ren_control = 160
    go_to_local = 161
    local_lockout = 162


class UsbTmcStatus(enum.IntEnum):
    success = 1
    pending = 2
    failed = 0x80
    transfer_not_in_progress = 0x81
    split_not_in_progress = 0x82
    split_in_progress = 0x83


UsbTmcCapabilities = namedtuple("UsbTmcCapabilities", "usb488 ren_control trigger")


def find_tmc_devices(
    vendor=None, product=None, serial_number=None, custom_match=None, **kwargs
):
    """Find connected USBTMC devices. See usbutil.find_devices for more info."""

    def is_usbtmc(dev):
        if custom_match and not custom_match(dev):
            return False
        return bool(find_interfaces(dev, bInterfaceClass=0xFE, bInterfaceSubClass=3))

    return find_devices(vendor, product, serial_number, is_usbtmc, **kwargs)


class BulkOutMessage(object):
    """The Host uses the Bulk-OUT endpoint to send USBTMC command messages to
    the device.

    """

    @staticmethod
    def build_array(btag, eom, chunk):
        size = len(chunk)
        return (
            struct.pack("BBBx", MsgID.dev_dep_msg_out, btag, ~btag & 0xFF)
            + struct.pack("<LBxxx", size, eom)
            + chunk
            + b"\0" * ((4 - size) % 4)
        )


class BulkInMessage(
    namedtuple(
        "BulkInMessage",
        "msgid btag btaginverse transfer_size transfer_attributes data",
    )
):
    """The Host uses the Bulk-IN endpoint to read USBTMC response messages from
    the device.

    The Host must first send a USBTMC command message that expects a response
    before attempting to read a USBTMC response message.

    """

    @classmethod
    def from_bytes(cls, data):
        msgid, btag, btaginverse = struct.unpack_from("BBBx", data)
        if msgid != MsgID.dev_dep_msg_in:
            warnings.warn(
                "Unexpected MsgID format. Consider updating the device's firmware. "
                "See https://github.com/pyvisa/pyvisa-py/issues/20"
                f"Expected message id was {MsgID.dev_dep_msg_in}, got {msgid}."
            )
            return BulkInMessage.from_quirky(data)

        transfer_size, transfer_attributes = struct.unpack_from("<LBxxx", data, 4)

        # Truncate data to the specified length (discard padding).
        data = data[12 : 12 + transfer_size]
        return cls(msgid, btag, btaginverse, transfer_size, transfer_attributes, data)

    @classmethod
    def from_quirky(cls, data):
        """Constructs a correct response for quirky devices."""
        msgid, btag, btaginverse = struct.unpack_from("BBBx", data)
        data = data.rstrip(b"\x00")
        # check whether it contains a ';' and if throw away the first 12 bytes
        if b";" in data:
            transfer_size, transfer_attributes = struct.unpack_from("<LBxxx", data, 4)
            data = data[12:]
        else:
            transfer_size = 0
            transfer_attributes = 1
        return cls(msgid, btag, btaginverse, transfer_size, transfer_attributes, data)

    @staticmethod
    def build_array(btag, transfer_size, term_char=None):
        """

        :param transfer_size:
        :param btag:
        :param term_char:
        :return:
        """

        if term_char is None:
            transfer_attributes = 0
            term_char = 0
        else:
            transfer_attributes = 2

        return struct.pack(
            "BBBx", MsgID.request_dev_dep_msg_in, btag, ~btag & 0xFF
        ) + struct.pack("<LBBxx", transfer_size, transfer_attributes, term_char)


class USBRaw(object):
    """Base class for drivers that communicate with instruments
    via usb port using pyUSB
    """

    #: Configuration number to be used. If None, the default will be used.
    CONFIGURATION = None

    #: Interface index it be used
    INTERFACE = (0, 0)

    #: Receive and Send endpoints to be used. If None the first IN (or OUT)
    #: BULK endpoint will be used.
    ENDPOINTS = (None, None)

    find_devices = staticmethod(find_devices)

    def __init__(
        self,
        vendor=None,
        product=None,
        serial_number=None,
        device_filters=None,
        timeout=None,
        **kwargs,
    ):
        super(USBRaw, self).__init__()

        # Timeout expressed in ms as an integer and limited to 2**32-1
        # If left to None pyusb will use its default value
        self.timeout = timeout

        device_filters = device_filters or {}
        devices = list(
            self.find_devices(vendor, product, serial_number, None, **device_filters)
        )

        if not devices:
            raise ValueError("No device found.")
        elif len(devices) > 1:
            desc = "\n".join(str(dev) for dev in devices)
            raise ValueError(
                f"{len(devices)} devices found:\n{desc}\nPlease narrow the search"
                " criteria"
            )

        self.usb_dev = devices[0]

        try:
            if self.usb_dev.is_kernel_driver_active(0):
                self.usb_dev.detach_kernel_driver(0)
        except (usb.core.USBError, NotImplementedError):
            pass

        try:
            cfg = self.usb_dev.get_active_configuration()
        except usb.core.USBError:
            cfg = None

        if cfg is None:
            try:
                self.usb_dev.set_configuration()
                cfg = self.usb_dev.get_active_configuration()
            except usb.core.USBError as e:
                raise Exception("failed to set configuration\n %s" % e)

        intf = cfg[(0, 0)]

        # Check if the interface exposes multiple alternative setting and
        # set one only if there is more than one.
        if (
            len(
                tuple(
                    usb.util.find_descriptor(
                        cfg, find_all=True, bInterfaceNumber=intf.bInterfaceNumber
                    )
                )
            )
            > 1
        ):
            try:
                self.usb_dev.set_interface_altsetting()
            except usb.core.USBError:
                pass

        self.usb_intf = self._find_interface(self.usb_dev, self.INTERFACE)

        self.usb_recv_ep, self.usb_send_ep = self._find_endpoints(
            self.usb_intf, self.ENDPOINTS
        )

    def _find_interface(self, dev, setting):
        return self.usb_dev.get_active_configuration()[self.INTERFACE]

    def _find_endpoints(self, interface, setting):
        recv, send = setting
        if recv is None:
            recv = find_endpoint(interface, usb.ENDPOINT_IN, usb.ENDPOINT_TYPE_BULK)
        else:
            recv = usb_find_desc(interface, bEndpointAddress=recv)

        if send is None:
            send = find_endpoint(interface, usb.ENDPOINT_OUT, usb.ENDPOINT_TYPE_BULK)
        else:
            send = usb_find_desc(interface, bEndpointAddress=send)

        return recv, send

    def write(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument
        :type data: bytes
        """

        try:
            return self.usb_send_ep.write(data)
        except usb.core.USBError as e:
            raise ValueError(str(e))

    def read(self, size):
        """Receive raw bytes to the instrument.

        :param size: number of bytes to receive
        :return: received bytes
        :return type: bytes
        """

        if size <= 0:
            size = 1

        data = self.usb_recv_ep.read(size, self.timeout).tobytes()

        return data

    def close(self):
        return usb.util.dispose_resources(self.usb_dev)


class USBTMC(USBRaw):
    find_devices = staticmethod(find_tmc_devices)

    def __init__(self, vendor=None, product=None, serial_number=None, **kwargs):
        super(USBTMC, self).__init__(vendor, product, serial_number, **kwargs)
        self.usb_intr_in = find_endpoint(
            self.usb_intf, usb.ENDPOINT_IN, usb.ENDPOINT_TYPE_INTERRUPT
        )

        time.sleep(0.01)

        self._capabilities = self._get_capabilities()

        self._btag = 0

        if not (self.usb_recv_ep and self.usb_send_ep):
            msg = "TMC device must have both Bulk-In and Bulk-out endpoints."
            raise ValueError(msg)

        self._enable_remote_control()

    def _enable_remote_control(self):
        if not self._capabilities.ren_control:
            return

        self.usb_dev.ctrl_transfer(
            usb.util.build_request_type(
                usb.util.CTRL_IN,
                usb.util.CTRL_TYPE_CLASS,
                usb.util.CTRL_RECIPIENT_INTERFACE,
            ),
            Request.ren_control,
            1,
            self.usb_intf.index,
            1,
            timeout=self.timeout,
        )

    def _get_capabilities(self):
        c = self.usb_dev.ctrl_transfer(
            usb.util.build_request_type(
                usb.util.CTRL_IN,
                usb.util.CTRL_TYPE_CLASS,
                usb.util.CTRL_RECIPIENT_INTERFACE,
            ),
            Request.get_capabilities,
            0x0000,
            self.usb_intf.index,
            0x0018,
            timeout=self.timeout,
        )

        usb488_capabilities = c[0xE]

        # bit #2: The interface is a 488.2 USB488 interface.
        # bit #1: The interface accepts REN_CONTROL, GO_TO_LOCAL,
        #         and LOCAL_LOCKOUT requests.
        # bit #0: The interface accepts the MsgID = TRIGGER
        #         USBTMC command message and forwards
        #         TRIGGER requests to the Function Layer.
        return UsbTmcCapabilities(
            usb488=bool(usb488_capabilities & (1 << 2)),
            ren_control=bool(usb488_capabilities & (1 << 1)),
            trigger=bool(usb488_capabilities & (1 << 0)),
        )

    def _find_interface(self, dev, setting):
        interfaces = find_interfaces(dev, bInterfaceClass=0xFE, bInterfaceSubClass=3)
        if not interfaces:
            raise ValueError("USB TMC interface not found.")
        elif len(interfaces) > 1:
            pass

        return interfaces[0]

    def _abort_bulk_in(self, btag):
        """Request that the device abort a pending Bulk-IN operation."""

        abort_timeout_ms = 5000

        # Send INITIATE_ABORT_BULK_IN.
        # According to USBTMC 1.00 4.2.1.4:
        #   wValue = bTag value of transfer to be aborted
        #   wIndex = Bulk-IN endpoint
        #   wLength = 0x0002 (length of device response)
        data = self.usb_dev.ctrl_transfer(
            usb.util.build_request_type(
                usb.util.CTRL_IN,
                usb.util.CTRL_TYPE_CLASS,
                usb.util.CTRL_RECIPIENT_ENDPOINT,
            ),
            Request.initiate_abort_bulk_in,
            btag,
            self.usb_recv_ep.bEndpointAddress,
            0x0002,
            timeout=abort_timeout_ms,
        )

        if data[0] != UsbTmcStatus.success:
            # Abort Bulk-IN failed. Ignore it.
            return

        # Read remaining data from Bulk-IN endpoint.
        self.usb_recv_ep.read(self.usb_recv_ep.wMaxPacketSize, abort_timeout_ms)

        # Send CHECK_ABORT_BULK_IN_STATUS until it completes.
        # According to USBTMC 1.00 4.2.1.5:
        #   wValue = 0x0000
        #   wIndex = Bulk-IN endpoint
        #   wLength = 0x0008 (length of device response)
        for retry in range(100):
            data = self.usb_dev.ctrl_transfer(
                usb.util.build_request_type(
                    usb.util.CTRL_IN,
                    usb.util.CTRL_TYPE_CLASS,
                    usb.util.CTRL_RECIPIENT_ENDPOINT,
                ),
                Request.check_abort_bulk_in_status,
                0x0000,
                self.usb_recv_ep.bEndpointAddress,
                0x0008,
                timeout=abort_timeout_ms,
            )
            if data[0] != UsbTmcStatus.pending:
                break
            time.sleep(0.05)

    def write(self, data):
        """Send raw bytes to the instrument.

        :param data: bytes to be sent to the instrument
        :type data: bytes
        """

        begin, end, size = 0, 0, len(data)
        bytes_sent = 0

        raw_write = super(USBTMC, self).write

        # Send all data via one or more Bulk-OUT transfers.
        # Set the EOM flag on the last transfer only.
        # Send at least one transfer (possibly empty).
        while (end == 0) or (end < size):
            begin, end = end, begin + self.usb_send_ep.wMaxPacketSize

            self._btag = (self._btag % 255) + 1

            eom = end >= size
            data = BulkOutMessage.build_array(self._btag, eom, data[begin:end])

            bytes_sent += raw_write(data)

        return size

    def read(self, size):
        header_size = 12
        max_padding = 511
        recv_chunk = self.usb_recv_ep.wMaxPacketSize - header_size

        if size > 0 and size < recv_chunk:
            recv_chunk = size

        eom = False

        raw_read = super(USBTMC, self).read
        raw_write = super(USBTMC, self).write

        received = bytearray()

        while not eom:
            self._btag = (self._btag % 255) + 1

            req = BulkInMessage.build_array(self._btag, recv_chunk, None)

            raw_write(req)

            try:
                resp = raw_read(recv_chunk + header_size + max_padding)
                response = BulkInMessage.from_bytes(resp)
            except (usb.core.USBError, ValueError):
                # Abort failed Bulk-IN operation.
                self._abort_bulk_in(self._btag)
                raise

            received.extend(response.data)

            # Detect EOM only when device sends all expected bytes.
            if len(response.data) >= response.transfer_size:
                eom = response.transfer_attributes & 1

        return bytes(received)
